/*   asl_functions.h various functions for the manipulation of ASL data

    Michael Chappell - FMRIB Image Analysis Group

    Moss Zhao - IBME Quantitative Biomedical Inference (QuBIc) Group

    Copyright (C) 2015 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#if !defined(asl_functions_h)
#define asl_functions_h

#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>

#include "armawrap/newmat.h"
#include "miscmaths/miscmaths.h"
#include "newimage/newimageall.h"

namespace OXASL
{
void data2stdform(NEWMAT::Matrix &datamtx, std::vector<NEWMAT::Matrix> &asldata, int ntis, std::vector<int> nrpts, bool isblocked, bool ispairs, bool blockpairs);
void stdform2data(std::vector<NEWMAT::Matrix> &asldata, NEWMAT::Matrix &datareturn, bool outblocked, bool outpairs);

// separate the pairs in the data (into seprate standard form items)
void separatepairs(std::vector<NEWMAT::Matrix> &asldata, std::vector<NEWMAT::Matrix> &asldataodd, std::vector<NEWMAT::Matrix> &asldataeven);
void mergepairs(std::vector<NEWMAT::Matrix> &asldata, std::vector<NEWMAT::Matrix> &asldataodd, std::vector<NEWMAT::Matrix> &asldataeven);

// mean the data at each TI
void timeans(std::vector<NEWMAT::Matrix> &asldata, std::vector<NEWMAT::Matrix> &aslmean);
// output data having taken mean at each TI
void timeanout(std::vector<NEWMAT::Matrix> &asldata, NEWIMAGE::volume<float> &mask, std::string fname, bool outpairs);

// output data split into individual TIs
void splitout(std::vector<NEWMAT::Matrix> &asldata, NEWIMAGE::volume<float> &mask, std::string froot);

// output epochs of data
void epochout(std::vector<NEWMAT::Matrix> &asldata, NEWIMAGE::volume<float> &mask, std::string froot, int epadv, int epol, bool outpairs, bool tiunit = false);
// generate epochs
void genepochs(std::vector<NEWMAT::Matrix> &asldata, std::vector<NEWMAT::Matrix> &epochreturn, int epadv, int epol);
// generate TI epochs
void gentiepochs(NEWMAT::Matrix &asldata, std::vector<NEWMAT::Matrix> &epochreturn, int epadv, int epol);

//output the result of deconvolution with an AIF
void deconvout(std::vector<NEWMAT::Matrix> &asldata, NEWIMAGE::volume<float> &mask, NEWMAT::Matrix &aif, std::string fname);
// do SVD convoloution
NEWMAT::ReturnMatrix SVDdeconv(const NEWMAT::Matrix &data, const NEWMAT::Matrix &aif);
// create a (simple) convolution matrix
NEWMAT::ReturnMatrix convmtx(const NEWMAT::ColumnVector &invec);

// Partial volume correction functions
// function to perform partial volume correction by linear regression
void pvcorr_LR(std::vector<NEWMAT::Matrix> &data_in, int ndata_in, NEWIMAGE::volume<float> &mask, NEWIMAGE::volume<float> &pv_map_gm, NEWIMAGE::volume<float> &pv_map_wm, int kernel, std::vector<NEWMAT::Matrix> &data_out, bool outblocked, bool outpairs, std::vector<int> nrpts, bool isblocked, bool ispairs, bool blockpairs);

// PV correction using linear regression (Asllani's method)
NEWIMAGE::volume<float> correct_pv_lr(const NEWIMAGE::volume<float> &data_in, const NEWIMAGE::volume<float> &mask, const NEWIMAGE::volume<float> &pv_map_gm, const NEWIMAGE::volume<float> &pv_map_wm, int kernel);

// Function to correct NaN values
NEWIMAGE::volume<float> correct_NaN(const NEWIMAGE::volume<float> &data_in);

// Extrapolation functions
// function to extrapolate voxels
void extrapolate(std::vector<NEWMAT::Matrix> &data_in, int ndata_in, int ntis, NEWIMAGE::volume<float> &mask, int neighbour_size, std::vector<NEWMAT::Matrix> &data_out, bool outblocked, bool outpairs, std::vector<int> nrpts, bool isblocked, bool ispairs, bool blockpairs);

// Function to do spiral search on 2D images and extrapolate edge voxels
NEWMAT::Matrix extrapolate_avg(NEWMAT::Matrix data_in, NEWMAT::Matrix mask, int neighbour_size);
}

#endif
